// NBody.cpp : Defines the entry point for the console application.
//

#include <iostream>
#include <vector>
#include <GL/freeglut.h>

#include "cl.hpp"
#include "Common.h"

bool keysPressed[256];
int visualizationWidth = 512;
int visualizationHeight = 512;

// visualization buffers
size_t visualizationBufferSize[2];
cl_float4* visualizationBufferCPU = NULL;
cl::Buffer visualizationBufferGPU;

// simulation buffer
const size_t bodyNum = 100;
// position + velocity
cl::Buffer particlesBuffer;

// kernels
std::vector<cl::Platform> platforms;
cl::Context context;
cl::CommandQueue queue;
std::vector<cl::Device> devices;
cl::Program program;
cl::Kernel visualizationClearKernel;
cl::Kernel visualizationKernel;

void resetSimulation();

void allocateVisualizationBuffers()
{
	visualizationBufferSize[0] = visualizationWidth;
	visualizationBufferSize[1] = visualizationHeight;
	if (visualizationBufferCPU != NULL)
	{
		delete[] visualizationBufferCPU;
	}
	visualizationBufferCPU = new cl_float4[visualizationWidth * visualizationHeight];
	cl_int err = CL_SUCCESS;
	visualizationBufferGPU = cl::Buffer(context, CL_MEM_READ_WRITE, sizeof(cl_float4) * visualizationWidth * visualizationHeight, NULL, &err);
}

int initSimulation()
{
	cl_int err = CL_SUCCESS;

	cl::Platform::get(&platforms);
	if (platforms.size() == 0)
	{
		std::cout << "Unable to find suitable platform." << std::endl;
		return -1;
	}

	cl_context_properties properties[] =
	{ CL_CONTEXT_PLATFORM, (cl_context_properties)(platforms[0])(), 0 }; // change back!
	context = cl::Context(CL_DEVICE_TYPE_GPU, properties, NULL, NULL, &err);
	devices = context.getInfo<CL_CONTEXT_DEVICES>(&err);
	queue = cl::CommandQueue(context, devices[0], 0, &err);

	std::string programSource = FileToString("..\\kernels\\programs.cl");
	program = cl::Program(context, programSource);
	err = program.build(devices);

	visualizationClearKernel = cl::Kernel(program, "visualizationClear", &err);
	visualizationKernel = cl::Kernel(program, "visualization", &err);

	allocateVisualizationBuffers();

	//TODO create simulation kernel
	particlesBuffer = cl::Buffer(context, CL_MEM_READ_WRITE, sizeof(cl_float4) * bodyNum, NULL, &err);

	resetSimulation();

	return 0;
}

void resetSimulation() {
	//TODO
	// fill particlesbuffer with random positions and velocities
}

void simulationStep() {
	//TODO
	// make one simulation step, which updates particlesBuffer
}

void visualizationStep()
{
	cl::Event event;
	cl_int err = CL_SUCCESS;
	visualizationClearKernel.setArg(0, visualizationWidth);
	visualizationClearKernel.setArg(1, visualizationHeight);
	visualizationClearKernel.setArg(2, visualizationBufferGPU);
	err = queue.enqueueNDRangeKernel(visualizationClearKernel, cl::NullRange, cl::NDRange(visualizationWidth, visualizationHeight), cl::NullRange, NULL, &event);

	visualizationKernel.setArg(0, visualizationWidth);
	visualizationKernel.setArg(1, visualizationHeight);
	visualizationKernel.setArg(2, visualizationBufferGPU);
	visualizationKernel.setArg(3, particlesBuffer);
	queue.enqueueNDRangeKernel(visualizationKernel, cl::NullRange, cl::NDRange(bodyNum), cl::NullRange, NULL, &event);

	err = queue.enqueueReadBuffer(visualizationBufferGPU, CL_TRUE, 0, sizeof(cl_float4) * visualizationWidth * visualizationHeight, visualizationBufferCPU, NULL, &event);

	glDrawPixels(visualizationWidth, visualizationHeight, GL_RGBA, GL_FLOAT, visualizationBufferCPU);
}

void initOpenGL()
{
	glClearColor(0.17f, 0.4f, 0.6f, 1.0f);
}

void display() {
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glDisable(GL_DEPTH_TEST);

	simulationStep();
	visualizationStep();

	glEnable(GL_DEPTH_TEST);
	glutSwapBuffers();
}

void idle() {
	glutPostRedisplay();
}

void keyDown(unsigned char key, int x, int y) {
	keysPressed[key] = true;
}

void keyUp(unsigned char key, int x, int y) {
	keysPressed[key] = false;
	switch (key) {

	case 'r':
		resetSimulation();
		break;

	case 27:
		exit(0);
		break;
	}
}

void mouseClick(int button, int state, int x, int y) {

}

void mouseMove(int x, int y) {

}

void reshape(int newWidth, int newHeight) {
	visualizationWidth = newWidth;
	visualizationHeight = newHeight;
	allocateVisualizationBuffers();
	glViewport(0, 0, visualizationWidth, visualizationHeight);
}

int main(int argc, char* argv[])
{
	glutInit(&argc, argv);
	glutInitContextVersion(3, 0);
	glutInitContextFlags(GLUT_CORE_PROFILE | GLUT_DEBUG);
	glutInitDisplayMode(GLUT_RGBA | GLUT_DEPTH | GLUT_DOUBLE);
	glutInitWindowSize(visualizationWidth, visualizationHeight);
	glutCreateWindow("NBODY");

	initOpenGL();

	glutDisplayFunc(display);
	glutIdleFunc(idle);
	glutReshapeFunc(reshape);
	glutKeyboardFunc(keyDown);
	glutKeyboardUpFunc(keyUp);
	glutMouseFunc(mouseClick);
	glutMotionFunc(mouseMove);

	// OpenCL processing
	initSimulation();

	glutMainLoop();
	return(0);
}

